package handlers_test

import (
	"bytes"
	"encoding/json"
	"errors"
	"net/http"
	"net/http/httptest"

	"github.com/cloudfoundry-incubator/notifications/fakes"
	"github.com/cloudfoundry-incubator/notifications/models"
	"github.com/cloudfoundry-incubator/notifications/web/handlers"
	"github.com/cloudfoundry-incubator/notifications/web/params"
	"github.com/ryanmoran/stack"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("CreateTemplate", func() {
	var (
		err         error
		handler     handlers.CreateTemplate
		writer      *httptest.ResponseRecorder
		request     *http.Request
		context     stack.Context
		creator     *fakes.TemplateCreator
		errorWriter *fakes.ErrorWriter
		database    *fakes.Database
	)

	Describe("ServeHTTP", func() {
		BeforeEach(func() {
			creator = fakes.NewTemplateCreator()
			errorWriter = fakes.NewErrorWriter()
			handler = handlers.NewCreateTemplate(creator, errorWriter)
			writer = httptest.NewRecorder()
			body := bytes.NewBuffer([]byte{})
			err := json.NewEncoder(body).Encode(map[string]interface{}{
				"name":    "Emergency Template",
				"text":    "Message to: {{.To}}. Raptor Alert.",
				"html":    "<p>{{.ClientID}} you should run.</p>",
				"subject": "Raptor Containment Unit Breached",
			})
			Expect(err).NotTo(HaveOccurred())

			database = fakes.NewDatabase()
			context = stack.NewContext()
			context.Set("database", database)

			request, err = http.NewRequest("POST", "/templates", body)
			Expect(err).NotTo(HaveOccurred())
		})

		It("calls create on its Creator with the correct arguments", func() {
			handler.ServeHTTP(writer, request, context)
			body := string(writer.Body.Bytes())

			Expect(creator.CreateCall.Arguments).To(ConsistOf([]interface{}{
				database,
				models.Template{
					Name:     "Emergency Template",
					Text:     "Message to: {{.To}}. Raptor Alert.",
					HTML:     "<p>{{.ClientID}} you should run.</p>",
					Subject:  "Raptor Containment Unit Breached",
					Metadata: "{}",
				},
			}))
			Expect(writer.Code).To(Equal(http.StatusCreated))
			Expect(body).To(Equal(`{"template_id":"guid"}`))
		})

		Context("when an errors occurs", func() {
			It("Writes a validation error to the errorwriter when the request is missing the name field", func() {
				body := []byte(`{"html": "<p>gobble</p>"}`)
				request, err = http.NewRequest("POST", "/templates", bytes.NewBuffer(body))
				if err != nil {
					panic(err)
				}

				handler.ServeHTTP(writer, request, context)
				Expect(errorWriter.Error).To(BeAssignableToTypeOf(params.ValidationError([]string{})))
			})

			It("Writes a validation error to the errorwriter when the request is missing the html field", func() {
				body := []byte(`{"name": "gobble"}`)
				request, err = http.NewRequest("POST", "/templates", bytes.NewBuffer(body))
				if err != nil {
					panic(err)
				}
				handler.ServeHTTP(writer, request, context)
				Expect(errorWriter.Error).To(BeAssignableToTypeOf(params.ValidationError([]string{})))
			})

			It("writes a parse error for an invalid request", func() {
				body := []byte(`{"name":"foobar", "html": forgot to close the curly brace`)
				request, err = http.NewRequest("POST", "/templates", bytes.NewBuffer(body))
				if err != nil {
					panic(err)
				}
				handler.ServeHTTP(writer, request, context)
				Expect(errorWriter.Error).To(BeAssignableToTypeOf(params.ParseError{}))
			})

			It("returns a 500 for all other error cases", func() {
				creator.CreateCall.Error = errors.New("my new error")
				handler.ServeHTTP(writer, request, context)
				Expect(errorWriter.Error).To(BeAssignableToTypeOf(params.TemplateCreateError{}))
			})
		})
	})
})
